#include "stdafx.h"
#include "JSAPrivate.hpp"

namespace JSA
{
    CDictionary::CDictionary(const DictionaryOptions & opt,const unsigned char * pData,unsigned long long nDataLength) : m_udtOptions(opt),
        m_pData(pData),m_nDataLength(nDataLength),m_nDictPos(0),m_nWindowLength(0),m_nOldestPos(0),m_pHashNodeAllocator(0),
        m_pHashTable(0)
    {
        m_pHashNodeAllocator = new HashNodeAllocator(m_udtOptions);
        m_pHashTable = new CHashTable;
    }

    CDictionary::~CDictionary()
    {
        if (m_pHashNodeAllocator != 0)
        {
            delete m_pHashNodeAllocator;
        }

        if (m_pHashTable != 0)
        {
            delete m_pHashTable;
        }
    }

    bool CDictionary::GetDictionaryOptions(DictionaryOptions & rOptions) const
    {
        rOptions = m_udtOptions;

        return true;
    }

    bool CDictionary::FindMatch(MatchData & rMatchData) const
    {
        rMatchData = MatchData();

        const unsigned long long nDataLeft = m_nDataLength - m_nDictPos;

        if (nDataLeft != 0)
        {
            if (nDataLeft >= m_udtOptions.MinMatchLength)
            {
                const unsigned long long nMaxMatch1 = nDataLeft < m_udtOptions.MaxMatchLength ? nDataLeft : m_udtOptions.MaxMatchLength;

                const unsigned long long nHashCode = static_cast<unsigned long long>(m_pData[m_nDictPos]) |
                    (static_cast<unsigned long long>(m_pData[m_nDictPos + 1]) << 8);

                const CHashList & rHashList = m_pHashTable->GetHashList(nHashCode);

                const unsigned long long nNodeCount = rHashList.GetNodeCount();

                const HashNode * pHashNode = rHashList.GetFirstHash();
                const HashNode * pNextNode = 0;

                unsigned long long nMatchedLength = 2;

                for (unsigned long x = 0; x < nNodeCount; ++x)
                {
                    pNextNode = pHashNode->NextNode;

                    const unsigned long long nWindowLeft = m_nDictPos - pHashNode->Position;
                    const unsigned long long nMaxMatch2 = nWindowLeft < nMaxMatch1 ? nWindowLeft : nMaxMatch1;

                    if (nMaxMatch2 < m_udtOptions.MinMatchLength)
                    {
                        // encode a literal
                        rMatchData.LZMatchData = MatchData::LZLiteral;
                        rMatchData.LZLiteralByte = m_pData[m_nDictPos];
                        break;
                    }
                    else
                    {
                        // check to see if this match could be longer
                        if (m_pData[pHashNode->Position + nMatchedLength] == m_pData[m_nDictPos + nMatchedLength])
                        {
                            nMatchedLength = CompareData(&m_pData[pHashNode->Position],&m_pData[m_nDictPos],2,nMaxMatch2);

                            if (nMatchedLength >= rMatchData.LZMatchLength)
                            {
                                rMatchData.LZMatchData = MatchData::LZMatch;
                                rMatchData.DataPosition = pHashNode->Position;
                                rMatchData.LZWindowIndex = m_nDictPos - pHashNode->Position - m_udtOptions.MinMatchLength;
                                rMatchData.LZMatchLength = nMatchedLength;
                                rMatchData.LZMatchBytes = &m_pData[m_nDictPos];

                                if (nMatchedLength == nMaxMatch2)
                                {
                                    break;
                                }
                            }
                        }
                    }

                    pHashNode = pNextNode;
                }

                if (rMatchData.LZMatchData != MatchData::LZMatch && rMatchData.LZMatchData != MatchData::LZLiteral)
                {
                    // encode a literal
                    rMatchData.LZMatchData = MatchData::LZLiteral;
                    rMatchData.LZLiteralByte = m_pData[m_nDictPos];
                }
            }
            else
            {
                // encode a literal
                rMatchData.LZMatchData = MatchData::LZLiteral;
                rMatchData.LZLiteralByte = m_pData[m_nDictPos];
            }
        }
        else
        {
            rMatchData.LZMatchData = MatchData::LZDone;
        }

        return true;
    }

    bool CDictionary::Update(const MatchData & rMatchData)
    {
        const unsigned long long nDataLeft = m_nDataLength - m_nDictPos;
        unsigned long long nHashesToAdd = 0;
        unsigned long long nAdd = 0;

        if (rMatchData.LZMatchData == MatchData::LZMatch)
        {
            nHashesToAdd = rMatchData.LZMatchLength;
            nAdd = nHashesToAdd;
        }
        else if (rMatchData.LZMatchData == MatchData::LZLiteral)
        {
            nHashesToAdd = 1;
            nAdd = 1;
        }


        if (nHashesToAdd == nDataLeft)
        {
            --nHashesToAdd;
        }

        // add the hashes
        for (unsigned long long i = 0; i < nHashesToAdd; ++i)
        {
            const unsigned long long nHashCode = static_cast<unsigned long long>(m_pData[m_nDictPos + i]) |
                (static_cast<unsigned long long>(m_pData[m_nDictPos + i + 1]) << 8);

            HashNode * pNode = m_pHashNodeAllocator->GetNode();

            pNode->Position = m_nDictPos + i;
            pNode->DataPtr = &m_pData[pNode->Position];

            CHashList & rHashList = m_pHashTable->GetHashList(nHashCode);

            rHashList.Add(pNode,m_udtOptions.ChainLimit);
        }

        m_nDictPos += nAdd;
        m_nWindowLength = m_nDictPos < m_udtOptions.WindowSize ? m_nDictPos : m_udtOptions.WindowSize;
        m_nOldestPos = m_nDictPos - m_nWindowLength;

        return true;
    }

    bool CDictionary::Finalize()
    {
        return true;
    }

    unsigned long long CDictionary::GetDataLength() const
    {
        return m_nDataLength;
    }

    unsigned long long CDictionary::GetDataPosition() const
    {
        return m_nDictPos;
    }

    unsigned long long CDictionary::CompareData(const unsigned char * pDictData,const unsigned char * pData,
                                                unsigned long long nPreMatchedLength,unsigned long long nMaxMatchLength)
    {
        unsigned long long nMatchLength = nPreMatchedLength;

        for (; nMatchLength < nMaxMatchLength; ++nMatchLength)
        {
            if (pDictData[nMatchLength] != pData[nMatchLength])
            {
                break;
            }
        }

        return nMatchLength;
    }

    IDictionary * IDictionary::GetDefaultDictionary(const DictionaryOptions & opt,const unsigned char * pData,unsigned long long nDataLength)
    {
        IDictionary * pDict = new CDictionary(opt,pData,nDataLength);

        return pDict;
    }
}
